//
// Project : webgen
// File    : parser.cc
// Purpose : functions for parsing the description file
// License : GPL
//
// arch-tag: 2e77116d-1f3a-4310-ba21-b4cec3528d59
//

#include <stdio.h>
#include <ctype.h>

#include <iostream>
#include <string>
#include <map>

#include "parser.hh"

///////////////////////////////////////////////////
//
// defines
//
///////////////////////////////////////////////////

//
// test for a token
//

#define IS_TOKEN( s )  ( strcmp( value, s ) == 0 )

///////////////////////////////////////////////////
//
// variables
//
///////////////////////////////////////////////////

static FILE * file = NULL;

static char cur_line[1024];
static uint pos;
static uint rows = 0;

extern std::map< std::string, Menu * > nodes;

///////////////////////////////////////////////////
//
// functions
//
///////////////////////////////////////////////////

//
// open file
//
int
init_parser ( const char * filename )
{
    file = fopen( filename, "r" );

    if ( file == NULL )
    {
        std::cout << "(init_parser) could not open file \"" << filename << "\"" << std::endl;
        return -1;
    }// if

    return 0;
}

//
// for messages
//
static void
print_token ( uint token, std::ostream & os )
{
    switch ( token )
    {
    case TOK_NOTHING : os << "<nothing>"; break;
    case TOK_EOF     : os << "<end-of-file>"; break;
    case TOK_SBLOCK  : os << "<{>"; break;
    case TOK_EBLOCK  : os << "<}>"; break;
    case TOK_ECMD    : os << "<;>"; break;
        
    case TOK_NUM     : os << "<number>"; break;
    case TOK_STRING  : os << "<string>"; break;
    case TOK_EQUAL   : os << "<equal>"; break;
        
    case TOK_MENU    : os << "<menu>"; break;
    case TOK_NAME    : os << "<name>"; break;
    case TOK_ID      : os << "<id>"; break;
    case TOK_CONTENT : os << "<content>"; break;
    case TOK_HTML    : os << "<htmlfile>"; break;
    case TOK_TITLE   : os << "<title>"; break;
    case TOK_HEADLN  : os << "<headline>"; break;
    case TOK_HEADER  : os << "<header>"; break;
    case TOK_FOOTER  : os << "<footer>"; break;
    case TOK_CSS     : os << "<css>"; break;
    case TOK_ADDCSS  : os << "<addcss>"; break;

    default:
        os << "<unknown token>"; 
    }// switch
}

static void
print_position ( std::ostream & os )
{
    os << "(" << rows << ", " << pos+1 << ")";
}

//
// read new line from file
//
int
get_line ( char * buf, int n )
{
    if ( file == NULL )
        return -1;
    
    if ( fgets( buf, n, file ) == NULL )
        return -1;

    rows++;
    
    return 0;
}

//
// parse for a number
//
int
parse_num ( char * buffer )
{
    uint  i = 0;
    
    while ( cur_line[pos] != '\0' )
    {
        char  c = cur_line[pos];
        
        if ( isdigit( cur_line[pos] ) )
            buffer[i++] = cur_line[pos++];
        else if ( isspace( c ) || ( c == ';' ))
        {
            buffer[i] = '\0';
            return 0;
        }// if
        else
        {
            buffer[i] = '\0';
            return -1;
        }// else
    }// while
    
    buffer[i] = '\0';

    return 0;
}

//
// parse for a name
//
int
parse_name ( char * buffer )
{
    uint  i = 0;
    
    while ( cur_line[pos] != '\0' )
    {
        char  c = cur_line[pos];
        
        if ( isalpha( c ) || isdigit( c ) || ( c == '_' ) )
            buffer[i++] = cur_line[pos++];
        else if ( isspace( c ) )
        {
            buffer[i] = '\0';
            return 0;
        }// if
        else
        {
            buffer[i] = '\0';
            return -1;
        }// else
    }// while

    buffer[i] = '\0';

    return 0;
}

//
// parse for a quoted string
//
int
parse_quoted_string ( char * buffer )
{
    uint  i = 0;

    // overread initial quote
    pos++;
    
    while ( cur_line[pos] != '\0' )
    {
        char  c = cur_line[pos];

        switch ( c )
        {
        case '"' :
            pos++;
            buffer[i] = '\0';
            return 0;

        case '\0' :
        case '\n' :
            buffer[i] = '\0';
            return -1;

        default:
            buffer[i++] = cur_line[pos++];
        }// switch
    }// while
    
    buffer[i] = '\0';

    return -1;
}

//
// return next token in stream
//
uint
get_token ( char * value )
{
    bool  no_token_found = true;
    uint  token;

    //
    // parse until error occurs or a token was found
    //
    
    while ( no_token_found )
    {
        // read next line from stream if all input was eaten up
        if ( pos >= strlen( cur_line ) )
        {
            if ( get_line( cur_line, sizeof( cur_line ) ) < 0 )
                return TOK_EOF;

            pos = 0;
        }// if

        //
        // parse for a token
        //

        while ( pos < strlen( cur_line ) )
        {
            if ( isspace( cur_line[pos] ) )
                pos++;
            else if ( cur_line[pos] == '#' )
            {
                // forget rest of line
                cur_line[0] = '\0';
                break;
            }// if
            else
            {
                if ( cur_line[pos] == '=' )
                {
                    no_token_found = false;
                    token = TOK_EQUAL;
                    strcpy( value, "=" );
                    pos++;
                    break;
                }// if
                else if ( cur_line[pos] == '{' )
                {
                    no_token_found = false;
                    token = TOK_SBLOCK;
                    strcpy( value, "{" );
                    pos++;
                    break;
                }// if
                else if ( cur_line[pos] == '}' )
                {
                    no_token_found = false;
                    token = TOK_EBLOCK;
                    strcpy( value, "}" );
                    pos++;
                    break;
                }// if
                else if ( cur_line[pos] == ';' )
                {
                    no_token_found = false;
                    token = TOK_ECMD;
                    strcpy( value, ";" );
                    pos++;
                    break;
                }// if
                else if ( isdigit( cur_line[pos] ) )
                {
                    if ( parse_num( value ) != 0 )
                    {
                        std::cout << "invalid number at (" << rows << ", " << pos+1 << ")"
                                  << std::endl;
                        token = TOK_NOTHING;
                    }// if
                    else
                        token = TOK_NUM;
                    no_token_found = false;
                    break;
                }// if
                else if ( cur_line[pos] == '"' )
                {
                    // read quoted string from line
                    if ( parse_quoted_string( value ) != 0 )
                    {
                        std::cout << "invalid string at (" << rows << ", " << pos+1 << ")"
                                  << std::endl;
                        token = TOK_NOTHING;
                    }// if
                    else
                        token = TOK_STRING;
                    
                    no_token_found = false;
                    break;
                }// if
                else
                {
                    // read for a name
                    if ( parse_name( value ) != 0 )
                    {
                        std::cout << "invalid name at (" << rows << ", " << pos+1 << ")"
                                  << std::endl;
                        token = TOK_NOTHING;
                    }// if

                    if ( IS_TOKEN( "menu" ) )
                    {
                        token = TOK_MENU;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "tab" ) )
                    {
                        token = TOK_TAB;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "id" ) )
                    {
                        token = TOK_ID;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "name" ) )
                    {
                        token = TOK_NAME;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "content" ) )
                    {
                        token = TOK_CONTENT;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "htmlfile" ) )
                    {
                        token = TOK_HTML;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "title" ) )
                    {
                        token = TOK_TITLE;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "headline" ) )
                    {
                        token = TOK_HEADLN;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "header" ) )
                    {
                        token = TOK_HEADER;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "footer" ) )
                    {
                        token = TOK_FOOTER;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "css" ) )
                    {
                        token = TOK_CSS;
                        no_token_found = false;
                        break;
                    }// if
                    else if ( IS_TOKEN( "addcss" ) )
                    {
                        token = TOK_ADDCSS;
                        no_token_found = false;
                        break;
                    }// if
                    else
                    {
                        token = TOK_NOTHING;
                        no_token_found = false;
                        break;
                    }// else
                }// else
            }// else
        }// while
    }// while

    return token;
}

//
// parse for a specific token
//
int
parse_token ( uint token, char * value )
{
    uint tok = get_token( value );

    if ( tok != token )
    {
        std::cout << "found token ";
        print_token( tok, std::cout );
        std::cout << " instead of ";
        print_token( token, std::cout );
        std::cout << " at (" << rows << ", " << pos+1 << ")" << std::endl;
        return -1;
    }// if
    else
        return 0;
}

///////////////////////////////////////////////////////
//
// functions defining grammar
//

static void
parse_id ( char * val )
{
    char  tmp[256];
    
    if ( parse_token( TOK_EQUAL, tmp ) != 0 )
        exit( 0 );

    uint tok = get_token( val );

    if (( tok != TOK_NUM ) && ( tok != TOK_STRING ))
    {
        std::cout << "found ";
        print_token( tok, std::cout );
        std::cout << " instead of num or string at ";
        print_position( std::cout );
        std::cout << std::endl;
        exit( 0 );
    }// else

    if ( parse_token( TOK_ECMD, tmp ) != 0 )
        exit( 0 );
}

static void
parse_string ( char * val )
{
    char  tmp[256];
    
    if ( parse_token( TOK_EQUAL, tmp ) != 0 )
        exit( 0 );

    if ( parse_token( TOK_STRING, val ) != 0 )
        exit( 0 );

    if ( parse_token( TOK_ECMD, tmp ) != 0 )
        exit( 0 );
}

static void
parse_tab ( Menu * menu )
{
    char    val[256];
    uint    tok;
    bool    end = false;
    bool    has_id, has_name, has_content, has_html, has_css;

    has_id = has_name = has_content = has_html = has_css = false;
    
    if ( parse_token( TOK_SBLOCK, val ) != 0 )
        exit( 0 );
    
    while ( ! end )
    {
        tok = get_token( val );
        
        switch ( tok )
        {
        case TOK_EBLOCK :
            end = true;
            break;
            
        case TOK_ID :
            if ( has_id )
                std::cout << "[warning] tab has already an ID" << std::endl;
            parse_id( val );
            menu->id() = val;
            
            if (( nodes[val] != NULL ) && ( nodes[val] != menu ))
                std::cout << "[error] : already have a node with this id (" << val << ")"
                          << std::endl;
            
            nodes[ val ] = menu;
            
            has_id = true;
            break;
            
        case TOK_NAME :
            if ( has_name )
                std::cout << "[warning] tab has already a name" << std::endl;
            parse_string( val );
            menu->name() = val;
            has_name = true;
            break;
            
        case TOK_CONTENT :
            if ( has_content )
                std::cout << "[warning] tab has already a content" << std::endl;
            parse_string( val );
            menu->content() = val;
            has_content = true;
            break;
            
        case TOK_HTML :
            if ( has_html )
                std::cout << "[warning] tab has already a html-file" << std::endl;
            parse_string( val );
            menu->html_file() = val;
            has_html = true;
            break;
            
        case TOK_CSS :
            if ( has_css )
                std::cout << "[warning] tab has already a css-file" << std::endl;
            parse_string( val );
            menu->set_new_css( true );
            menu->css_files().append( val );
            has_css = true;
            break;
            
        case TOK_ADDCSS :
            parse_string( val );
            menu->css_files().append( val );
            break;
            
        case TOK_EOF :
            std::cout << "unexpected end of file" << std::endl;
            exit( 0 );
            
        default :
            std::cout << "unexpected token ";
            print_token( tok, std::cout );
            std::cout << " at ";
            print_position( std::cout );
            std::cout << std::endl;
            exit( 0 );
        }// switch
    }// while

    // check if menu is just a link
    if ( has_html && ! has_content )
        menu->set_link( true );

    // set html-file according to content-file
    if ((menu->html_file() == "") && (menu->content() != ""))
    {
        int  i = 0;
        bool has_dot = false;
        
        menu->html_file() = menu->content();
        
        // look for '.'
        for ( i = menu->html_file().length() - 1; i >= 0; i-- )
        {
            if ( menu->html_file()[i] == '.' )
            {
                has_dot = true;
                break;
            }// if
        }// for
        
        if ( has_dot )
            menu->html_file() = menu->html_file().substr( 0, i );
        
        menu->html_file() = menu->html_file() + ".html";
    }// if
}

static void
parse_menu ( Menu * menu )
{
    char    val[256];
    uint    tok;
    bool    end = false;
    bool    has_id, has_name, has_content, has_html;
    bool    has_title, has_headline, has_header, has_footer, has_css;
    Menu  * sub;

    has_id = has_name = has_content = has_html = false;
    has_title = has_headline = has_header = has_footer = has_css = false;
    
    if ( parse_token( TOK_SBLOCK, val ) != 0 )
        exit( 0 );
    
    while ( ! end )
    {
        tok = get_token( val );
        
        switch ( tok )
        {
        case TOK_EBLOCK :
            end = true;
            break;
            
        case TOK_ID :
            if ( has_id )
                std::cout << "[warning] menu has already an ID" << std::endl;
            parse_id( val );
            menu->id() = val;
            
            if (( nodes[val] != NULL ) && ( nodes[val] != menu ))
                std::cout << "[error] : already have a node with this id (" << val << ")"
                          << std::endl;
            
            nodes[ val ] = menu;
            
            has_id = true;
            break;
            
        case TOK_NAME :
            if ( has_name )
                std::cout << "[warning] menu has already a name" << std::endl;
            parse_string( val );
            menu->name() = val;
            has_name = true;
            break;
            
        case TOK_CONTENT :
            if ( has_content )
                std::cout << "[warning] menu has already a content" << std::endl;
            parse_string( val );
            menu->content() = val;
            has_content = true;
            break;
            
        case TOK_HTML :
            if ( has_html )
                std::cout << "[warning] menu has already a html-file" << std::endl;
            parse_string( val );
            menu->html_file() = val;
            has_html = true;
            break;
            
        case TOK_TITLE :
            if ( has_title )
                std::cout << "[warning] menu has already a title" << std::endl;
            parse_string( val );
            menu->title() = val;
            has_title = true;
            break;
            
        case TOK_HEADLN :
            if ( has_headline )
                std::cout << "[warning] menu has already a headline" << std::endl;
            parse_string( val );
            menu->headline() = val;
            has_headline = true;
            break;
            
        case TOK_HEADER :
            if ( has_header )
                std::cout << "[warning] menu has already a header-file" << std::endl;
            parse_string( val );
            menu->header() = val;
            has_header = true;
            break;
            
        case TOK_FOOTER :
            if ( has_footer )
                std::cout << "[warning] menu has already a footer-file" << std::endl;
            parse_string( val );
            menu->footer() = val;
            has_footer = true;
            break;
            
        case TOK_CSS :
            if ( has_css )
                std::cout << "[warning] menu has already a css-file" << std::endl;
            parse_string( val );
            menu->set_new_css( true );
            menu->css_files().append( val );
            has_css = true;
            break;
            
        case TOK_ADDCSS :
            parse_string( val );
            menu->css_files().append( val );
            break;
            
        case TOK_MENU :
            sub = new Menu( menu, "", "", "", "" );
            menu->add_submenu( sub );
            parse_menu( sub );
            break;
            
        case TOK_TAB :
            sub = new Menu( menu, "", "", "", "" );
            sub->set_is_tab( true );
            menu->add_tab( sub );
            parse_tab( sub );
            break;
            
        case TOK_EOF :
            std::cout << "unexpected end of file" << std::endl;
            exit( 0 );
            
        default :
            std::cout << "unexpected token ";
            print_token( tok, std::cout );
            std::cout << " at ";
            print_position( std::cout );
            std::cout << std::endl;
            exit( 0 );
        }// switch
    }// while

    // check if menu is just a link
    if ( has_html && ! has_content )
        menu->set_link( true );

    // set html-file according to content-file
    if ((menu->html_file() == "") && (menu->content() != ""))
    {
        int  i = 0;
        bool has_dot = false;
        
        menu->html_file() = menu->content();
        
        // look for '.'
        for ( i = menu->html_file().length() - 1; i >= 0; i-- )
        {
            if ( menu->html_file()[i] == '.' )
            {
                has_dot = true;
                break;
            }// if
        }// for
        
        if ( has_dot )
            menu->html_file() = menu->html_file().substr( 0, i );
        
        menu->html_file() = menu->html_file() + ".html";
    }// if
}

void
parse ( Menu * menu, std::string & menufile )
{
    char  val[1024];
    
    init_parser( menufile.c_str() );

    if ( parse_token( TOK_MENU, val ) != 0 )
        exit( 0 );

    parse_menu( menu );
}

//////////////////////////////////////////////////////
//
// read menu from given file and root (obsolete)
//
#if 0
void
read_menu ( Menu & root, const string & filename )
{
    FILE  * file;

    if ((file = fopen( filename.c_str(), "r" )) == NULL)
    {
        cout << "read_menu : could not open " << filename << endl;
        return;
    }// if

    //
    // read alll entries
    //

    char    buffer[1024];
    string  id, name, father, content, html_file, title, headline;
    string  header, footer, cssfile;
    TSLL< std::string > css_files;
    bool    is_link;
    
    while ( ! feof( file ) )
    {
        unsigned  i = 0;
        
        fgets( buffer, sizeof(buffer), file );

        if (buffer[ strlen(buffer) - 1 ] == '\n')
            buffer[ strlen(buffer) - 1 ] = 0;

        // remove leading whitespaces and comments
        while ( i < strlen( buffer ) )
        {
            if ( buffer[i] == ' ' )
            {
                memmove( buffer + i, buffer + i + 1, strlen(buffer) - i );
                continue;
            }// if
            else if ( buffer[i] == '#' )
            {
                buffer[i] = 0;
                break;
            }// if
            else
                break;

            i++;
        }// for
        
        if (strlen(buffer) == 0)
            continue;

        id        = next_entry( buffer, sizeof(buffer) );
        name      = next_entry( buffer, sizeof(buffer) );
        father    = next_entry( buffer, sizeof(buffer) );
        content   = next_entry( buffer, sizeof(buffer) );
        html_file = next_entry( buffer, sizeof(buffer) );
        title     = next_entry( buffer, sizeof(buffer) );
        headline  = next_entry( buffer, sizeof(buffer) );
        header    = next_entry( buffer, sizeof(buffer) );
        footer    = next_entry( buffer, sizeof(buffer) );

        css_files.remove_all();
        
        while (( cssfile = next_entry( buffer, sizeof(buffer) ) ) != "" )
            css_files.append( cssfile );

        // check for empty line
        if ( id == "" )
            continue;
            
        if ((content == "" ) && (html_file != ""))
            is_link = true;
        else
            is_link = false;
        
        if ((html_file == "") && (content != ""))
        {
            int  i = 0;
            bool has_dot = false;
            
            html_file = content;

            // look for '.'
            for ( i = html_file.length() - 1; i >= 0; i-- )
            {
                if ( html_file[i] == '.' )
                {
                    has_dot = true;
                    break;
                }// if
            }// for

            if ( has_dot )
                html_file = html_file.substr( 0, i );

            html_file = html_file + ".html";
        }// if
        
        if ( name == root.name() )
        {
            root.title()     = title;
            root.headline()  = headline;
            root.header()    = header;
            root.footer()    = footer;
            root.css_files().copy( css_files );

            if (verbose)
            {
                cout << "node = (" << root.name() << ")" << endl
                     << "  id        = (" << root.id() << ")" << endl
                     << "  content   = (" << root.content() << ")" << endl
                     << "  html_file = (" << root.html_file() << ")" << endl
                     << "  title     = (" << root.title() << ")" << endl
                     << "  headline  = (" << root.headline() << ")" << endl
                     << "  header    = (" << root.header() << ")" << endl
                     << "  footer    = (" << root.footer() << ")" << endl;

                if ( root.css_files().size() > 0 )
                {
                    cout << "  cssfiles  = (";

                    TSLL< std::string >::TIterator  iter = root.css_files().first();

                    while ( ! iter.eol() > 0 )
                    {
                        cout <<  iter() << ", ";
                        ++iter;
                    }// while
                    
                    cout << ")" << endl;
                }// if
            }// if
        }// if
        else
        {
            Menu * father_node = nodes[ father ];

            if ( father_node == NULL )
            {
                cout << "read_menu : unkown father node " << father << endl;
                continue;
            }// if
            
            Menu * new_node = new Menu( father_node, id, name, content, html_file );

            if ( is_link )
                new_node->set_link( true );
            
            if (title != "") new_node->title() = title;
            else             new_node->title() = father_node->title() + " : " + name;

            if ( headline != "" ) new_node->headline() = headline;
            else                  new_node->headline() = father_node->headline();

            if ( header != "" ) new_node->header() = header;
            else                new_node->header() = father_node->header();

            if ( footer != "" ) new_node->footer() = footer;
            else                new_node->footer() = father_node->footer();

            if ( css_files.size() > 0 ) new_node->css_files().copy( css_files );
            else                        new_node->css_files().copy( father_node->css_files() );

            if ( nodes[ id ] != NULL )
                cout << "read_menu : already have a node with this id (" << id << ")" << endl;
            
            nodes[ id ] = new_node;
            father_node->add_submenu( new_node );

            if (verbose)
            {
                cout << "node = (" << new_node->name() << ")" << endl
                     << "  father    = (" << new_node->father()->name() << ")" << endl
                     << "  id        = (" << new_node->id() << ")" << endl
                     << "  content   = (" << new_node->content() << ")" << endl
                     << "  html_file = (" << new_node->html_file() << ")" << endl
                     << "  title     = (" << new_node->title() << ")" << endl
                     << "  headline  = (" << new_node->headline() << ")" << endl
                     << "  header    = (" << new_node->header() << ")" << endl
                     << "  footer    = (" << new_node->footer() << ")" << endl;

                
                if ( new_node->css_files().size() > 0 )
                {
                    cout << "  cssfiles  = (";

                    TSLL< std::string >::TIterator  iter = new_node->css_files().first();

                    while ( ! iter.eol() > 0 )
                    {
                        cout <<  iter() << ", ";
                        ++iter;
                    }// while

                    cout << ")" << endl;
                }// if
            }// if
        }// else
    }// while

    // finish
    fclose( file );
}

//
// parse given string until comma apears and return prefix (strtok)
//
static string
next_entry ( char * buffer, int size )
{
    int     i = 0;
    string  tmp;

    while ((i < size) && (buffer[i] != 0) && (buffer[i] != ','))
    {
        tmp += buffer[i];
        i++;
    }// while

    unsigned int  j;

    for ( j = i+1; j < strlen(buffer); j++ )
        buffer[j-(i+1)] = buffer[j];

    if ( ((size_t)i+1) > strlen(buffer) ) buffer[ 0 ] = 0;
    else                                  buffer[ strlen(buffer) - (i+1) ] = 0;

    return tmp;
}

#endif
